"""
This module contains tools for developers contributing to Pyttman.

It helps with building the framework for testing and simplifies
deployment, and makes it easier to verify functionality of tools such
as Pyttman CLI and testing the framework using a local-only test-app
with the most recent build of the framework.
"""

